#version 150
precision highp float; precision highp int;
//Ethan Alexander Shulman 2022 "Concrete Skies" July 23, 2022

in vec3 in_Position;
out vec4 uv;

layout(std140) uniform ShaderviewBlock {
	vec4 mouse;
	float screenX,screenY;
	int frame,framerate;
};


//config
#define ITERATIONS 80
#define EPSILON 1e-2
#define EPSILON_SCALE 1.06
//#define DISTANCE_STEP 1.
#define STEP_DITHER .5
//#define VOLUME_EPSILON 1e-1
#define LIGHTING_EPSILON 1e-5
//#define MAX_BOUNCES 6.

#define LIGHT_IMPORTANCE .3
#define LIGHT_DIRECTION normalize(vec3(2,.8,.7))
#define LIGHT_SIZE .03
#define LIGHT_COLOR vec3(2)
#define LIGHT_RESCALE_MAX 1e3
#define LIGHT_CAUSTICS 10.
#define LIGHT_MASK(p) step(200.,p.y)

#define FOG(p) vec4(vec3(.99),clamp(1.-p.y/200.,0.,1.)*.02)
#define FOG_ROUGHNESS 1.
#define FOG_ISOTROPY 1.

#define ANIMATION_TIME 60.
#define ANIMATION_SAMPLES 100
#define ANIMATION_RENDER 1
out float time;


//globals
#define PI 3.141592653589793

#ifdef ANIMATION_TIME
#define SFRAME (frame%ANIMATION_SAMPLES)
#else
#define SFRAME frame
#endif


//utility functions
mat2 r2d(float a) {
	float c = cos(a), s = sin(a);
	return mat2(c,s,-s,c);
}

uvec4 hash(uvec4 seed) {
    uvec4 h = (0x6A7F8FAAu^seed)*0x01000193u;
    h = ((h.wxyz>>3u)^h^seed.yzwx)*0x01000193u;
    h = ((h.zwxy>>8u)^h^seed.wxyz)*0x01000193u;
    return h^(h>>11u);
}
#define I2F (1./float(0xFFFFFFFFu))


float noise(vec3 p) {
	p = floor(abs(p));
	return fract(dot(p,vec3(.3123456789,.456791234,.6987654321)));
}
#define LNBLUR(p) cos(p.zxy*87162.8976234)



//scene definition

/*returns camera data as mat3 in the following structure {
vec3 position;
float roll, pitch, yaw;
float fieldOfViewTangent, viewNear, orbitDistance;*/
mat3 camera() {
	/*animated bezier camera path*/
	const float CURVE_TICK = 10.;
	const mat3 path[] = mat3[](mat3(-23,120,-55, 0,.1,-.2, .8,0,0), mat3(-25,120,-55, 0,.4,.4, .8,0,0),
				mat3(0,120,-55, 0,.04,1.8, 1.,0,0), mat3(20,170,-80, 0,0.3,2.8, 1.,0,0),
				mat3(20,180,-90, 0,0.2,3.6, 1.,0,0), mat3(30,180,-90, 0,0.1,5., 1.,0,0),
				mat3(0,180,-90, 0,-.8,6., 1.,0,0), mat3(0,100,-90, 0,-1.2,6.1, 1.,0,0),
				mat3(0,20,70, 0,.2,6.05, 1.,0,0), mat3(0,40,90, 0,0,5.2, 1.,0,0),
				mat3(-40,70,120, 0,.05,4.8, 1.,0,0), mat3(-100,100,110, 0,0.04,4.2, .8,0,0),
				mat3(-100,100,110, 0,.1,4.2, .5,0,0));
	float t = time/CURVE_TICK, ptick = floor(t);
	t -= ptick;
	int ind = int(ptick*2.);
	float invt = 1.-t;
	return path[ind]*invt*invt+path[ind+1]*invt*t*2.+path[ind+2]*t*t;
			
/*screenshot 0
	return mat3(-6,175,0,
			0,-.4,.5,
			.75,0,0);*/
			
/*screenshot 1
	return mat3(-80,96,150,
			0,.1,3.9,
			.9,0,0);*/
}

//returns vec2(distance, materialId) of nearest surface
vec2 geometry(vec3 p) {
	//buildings
	vec2 geo = vec2(length(max(abs(mod(abs(p.xz),100.)-50.)-20.,0.)),0);
	
	//z axis bridges
	vec2 bridgeX = mod(abs(p.xy),vec2(100,20))-vec2(50,10);
	float brx = length(max(abs(bridgeX)-vec2(6.,.6),0.));
	if (brx < geo.x) geo.x = brx;
	bridgeX.y -= 4.02;
	bridgeX = abs(bridgeX)-vec2(3.,3.);
	if (max(bridgeX.x,bridgeX.y) <= 0.) geo.y = bridgeX.x<-2.9?2.:1.;

	//x axis bridges
	vec2 bridgeZ = mod(abs(p.zy),vec2(100,20))-vec2(50,10);
	float brz = length(max(abs(bridgeZ)-vec2(6.,.6),0.));
	if (brz < geo.x) geo.x = brz;
	bridgeZ.y -= 4.02;
	bridgeZ = abs(bridgeZ)-vec2(3.,3.);
	if (max(bridgeZ.x,bridgeZ.y) <= 0.) geo.y = bridgeZ.x<-2.9?2.:1.;
	
	geo.x = max(p.y-200.,geo.x);

	return geo;
}
vec2 geometryDetail(vec3 p) {
	vec2 geo = geometry(p);
	if (geo.y == 0.) {
		//concrete bumps
		p += cos(p.yzx*2.57834)*.1;
		p += cos(p.zxy*.98723);
		geo.x += length(mod(abs(p),.4)-.2)*.25;
	}
	return geo;
}

/*returns material as mat3 in the following structure {
vec3 color;
float roughness, emission, opacity;*/
mat2x3 material(vec3 p, float mat) {

	//glass doors
	if (mat != 0.) return mat2x3(vec3(.9), .002+(mat-1.)*.1,0,1);

	//concrete
	vec3 blur = LNBLUR(p);
	return mat2x3(vec3(.01+noise(p/4.+blur*.5)*.2+noise(p+blur*.5)*.2), .45,0,1);
}


//returns background color at point and direction
vec3 background(vec3 p, vec3 d) {
	return vec3(clamp(p.y/100.-1.,0.,1.));
}
vec3 backgroundLighting(vec3 p, vec3 d) {
	return background(p,d)*0.1;
}


//sample detailed normal using derivatives from geometryDetail
vec3 normal(vec3 p) {
	vec3 n = normalize(vec3(geometryDetail(p+vec3(LIGHTING_EPSILON,0,0)).x,geometryDetail(p+vec3(0,LIGHTING_EPSILON,0)).x,geometryDetail(p+vec3(0,0,LIGHTING_EPSILON)).x)-geometryDetail(p).x);
	if (n != n) return vec3(0,1,0);
	return n;
}


//bidirectional path tracer
uniform sampler2D Buffer0;

void vert() {
	#ifdef ANIMATION_TIME
	#ifdef ANIMATION_RENDER
	time = float(frame)/float(framerate*ANIMATION_SAMPLES);
	#else
	time = mod(float(frame)/float(framerate),ANIMATION_TIME);
	#endif
	#else
	time = 0.;
	#endif
	uv = vec4(in_Position.xy*vec2(screenX/screenY,1),(in_Position.xy*.5+.5)*vec2(screenX,screenY));
	gl_Position = vec4(in_Position,1);
}
void frag() {
	//setup camera ray
	mat3 cam = camera();
	vec4 random = hash(uvec4(SFRAME)*uvec4(1,10,100,1000))*I2F;
	vec3 rp = vec3(0,0,cam[2].z),
		rd = normalize(vec3(uv.xy+(random.xy*2.-1.)/screenY,cam[2].x));
	if (cam[1].x != 0.) {
		mat2 m = r2d(cam[1].x);
		rp.xy *= m; rd.xy *= m;
	}
	if (cam[1].y != 0.) {
		mat2 m = r2d(cam[1].y);
		rp.yz *= m; rd.yz *= m;
	}
	if (cam[1].z != 0.) {
		mat2 m = r2d(cam[1].z);
		rp.xz *= m; rd.xz *= m;
	}
	rp += cam[0].xyz+rd*cam[2].y;

	//path trace
	vec3 c = vec3(1);//ray color tint
	vec4 l = vec4(0);//l.xyz = light color accumulation, l.w = number of bounces or -1 if important ray
	float epsilon = EPSILON;
	for (int i = 0; i < ITERATIONS && rp.y < 200.; i++) {
		vec2 geo = geometry(rp);
		#ifdef FOG
		bool hit;
		vec4 fogV;
		if (geo.x > 0.) {
			fogV = FOG(rp);
			if (fogV.w <= 0.) hit = false;
			else {
				#ifdef STEP_DITHER
				float stepSz = (epsilon+geo.x)*((1.-STEP_DITHER)+random.w*STEP_DITHER);
				#else
				float stepSz = epsilon+geo.x;
				#endif
				if (pow(1.-fogV.w,stepSz) < random.w) {
					hit = true;
					fogV.w = 1.;
					rp += rd*stepSz;
				} else hit = false;
			}
		} else {
			hit = true;
			fogV.w = 0.;
		}
		if (hit) {
			mat2x3 mat;
			if (fogV.w == 1.) mat = mat2x3(fogV.xyz, FOG_ROUGHNESS,0,1);//fog material
			else mat = material(rp,geo.y);//sample material
		
		#else
		if (geo.x <= 0.) {//hit/inside geometry
			mat2x3 mat = material(rp,geo.y);//sample material
		#endif

			#ifdef VOLUME_EPSILON
			if (random.w < mat[1].z) {//stochastic opacity
			#endif
			
				#ifdef LIGHT_IMPORTANCE
				if (l.w < 0.) {	//failed importance cast
					c = vec3(0);
					break;
				}
				#endif
			
				rp -= rd*epsilon*2.;//back up ray
				c *= mat[0];//apply material color
				#ifdef LIGHT_IMPORTANCE
				l += vec4(c.xyz*mat[1].y/max(1./LIGHT_RESCALE_MAX,pow(1.-LIGHT_IMPORTANCE,l.w)),1);//emissive lighting + bounce count
				#else
				l += vec4(c.xyz*mat[1].y,1);//emissive lighting + bounce count
				#endif
				#ifdef MAX_BOUNCES
				if (l.w >= MAX_BOUNCES) break;
				#endif
				random = hash(uvec4(abs(rp.xyz)*10000.,SFRAME))*I2F;
				#ifdef FOG
				#ifdef FOG_ISOTROPY
				vec3 nrm = fogV.w==1.?normalize((random.xyz*2.-1.)*FOG_ISOTROPY-rd):normal(rp);
				#else
				vec3 nrm = fogV.w==1.?-rd:normal(rp);
				#endif
				#else
				vec3 nrm = normal(rp);
				#endif

				#ifdef LIGHT_IMPORTANCE
				if (random.w < LIGHT_IMPORTANCE) {//directional light importance sampling
					//brdf
					float edge = (1.-LIGHT_SIZE*.1)-mat[1].x*4.,
						brdf = clamp((dot(LIGHT_DIRECTION,reflect(rd,nrm))-edge)/(1.-min(1.,edge)),0.,1.);
					rd = normalize(LIGHT_DIRECTION+(random.xyz-.5)*LIGHT_SIZE);
					
					if (dot(rd,nrm) < -.2 || brdf < 0.) {
						c = vec3(0);
						break;
					}
					c *= brdf/(1.+mat[1].x*LIGHT_CAUSTICS);
					if (l.w > 1.) c *= 1./max(1./LIGHT_RESCALE_MAX,pow(1.-LIGHT_IMPORTANCE,l.w-1.));//rescale brightness of bounces
					l.w = -1.;//mark ray important
					
				} else {
				#endif
				
				//randomly bounce ray off microsurface bump
				vec3 tang = normalize(cross(abs(nrm.y)>(1.-1e-5)?vec3(0,0,-1):vec3(0,-1,0),nrm)), bitang = cross(tang,nrm);
				float cs = pow(random.x,mat[1].x), sn = sqrt(1.-cs*cs), ang = random.y*PI*2.;
				rd = reflect(rd,normalize(mat3(nrm,tang,bitang)*vec3(cs,sn*cos(ang),sn*sin(ang))));
				#ifdef LIGHT_IMPORTANCE
				}
				#endif
				
				geo.x = 0.;
			#ifdef VOLUME_EPSILON
			} else {
				geo.x = VOLUME_EPSILON;
			}
			#endif
		} else {
			#ifdef DISTANCE_STEP
			geo.x *= DISTANCE_STEP;
			#endif
		}
		
		//ray march
#ifdef STEP_DITHER
		float stepSz = (epsilon+geo.x)*((1.-STEP_DITHER)+random.w*STEP_DITHER);
#else
		float stepSz = epsilon+geo.x;
#endif
		rp += rd*stepSz;
		random.w = fract(random.w+stepSz*2763.861234);
#ifdef EPSILON_SCALE
		epsilon *= EPSILON_SCALE;
#endif
	}
	
	#ifdef LIGHT_IMPORTANCE
	if (l.w < 0.) {
		l.xyz += c*LIGHT_COLOR*(1.+LIGHT_CAUSTICS)*LIGHT_MASK(rp)/LIGHT_IMPORTANCE;
	} else {
		if (l.w > 0.) c *= 1./max(1./LIGHT_RESCALE_MAX,pow(1.-LIGHT_IMPORTANCE,l.w));//rescale brightness of bounces
	#endif
	//background
	l.xyz += c*(l.w>0.?backgroundLighting(rp,rd):background(rp,rd));
	#ifdef LIGHT_IMPORTANCE
	}
	#endif
	
	//sum light in buffer
	vec4 sum = texelFetch(Buffer0,ivec2(uv.zw),0);
	if (SFRAME == 0) sum = vec4(0);
	gl_FragColor = sum+vec4(l.xyz,1);
}
